# 常量和变量

常量和变量是程序中所要操作的数据的载体。常量和变量的区别其实主要在于其中所承载的数据是否可变，数据的可变性决定了承载指代它们的容器，并且决定了它们在程序中能够发挥的作用。

## 常量

常量顾名思义就是不可变的量，在程序中定义的常量一般都携带有特殊的含义。根据常量类型的不同，一般可以按参照类型分为布尔常量、字符常量、整数常量、浮点常量、复数常量和字符串常量，其中字符常量、整数常量、浮点常量和复数常量一般统称为数字常量。

常量主要是通过字面量、其他的常量标识符、常量表达式等定义的，其定义原则就是用来定义常量的值在 Go 项目编译的时候是能够确定的，只要是能够符合这个条件的值，都可以用来定义常量。常量可以是类型化的也可以是非类型化的，像是字面量、`true`、`false`、`iota`或者是包含非类型化常量的常量表达式都是非类型化的。常量的类型可以在常量定义表达式中显式声明，也可以通过在赋值表达式中为变量赋值时隐式声明。

```admonish tip
隐式类型声明都是通过类型推断完成的。如果给定常量值不能匹配常量或者变量的类型，那么Go将产生一个错误。如果目标类型是一个类型参数，那么这个常量值将会被转换成匹配类型参数的非常量值。

非类型化的常量并不等于常量没有类型，非类型化只是指常量的类型是隐式声明的而已。
```

数字常量通常都标识任意精度的准确值，不会存在溢出的情况，所以不会存在负零、无穷大和非数字值的常量。尽管数字常量在 Go 中拥有任意精度，但是在编译器内它们的精度还是有限制的。

程序所使用的常量都是通过`const`关键字定义的，其语法定义如下：

```
ConstDecl = "const" ( ConstSpec | "(" { ConstSpec ";" } ")" ) .
ConstSpec = IdentifierList [ [ Type ] "=" ExpressionList ] .
IdentifierList = identifier { "," identifier } .
ExpressionList = Expression { "," Expression } .
```

以下是一些常见的常量定义，以及它们所定义出的内容类型。

```go
const a = 2 + 3.0 // a = 5.0，非类型化浮点常量
const b = 15 / 4 // b = 3，非类型化整型常量
const c = 15 / 4.0 // c = 3.75，非类型化浮点常量
const θ float64 = 3 / 2 // θ = 1.0，类型化双精度浮点常量，
const λ float64 = 3 / 2. // λ = 1.5，类型化双精度浮点常量，
const d = 1 << 3.0 // d = 8，非类型化整型常量
const e = 1.0 << 3 // e = 8，非类型化整型常量
const f = int32(1) << 33 // 非法常量，将常量值赋予int32类型常量是发生了溢出
const g = float64(2) >> 1 // 非法常量，float64(2)是一个类型化的常量
const h = "foo" > "bar" // h = true，非类型化布尔常量
const j = true // j = true，非类型化布尔常量
const k = 'w' + 1 // k = 'x'，非类型化字符常量
const l = "hi" // l = "hi"，非类型化字符串常量
const m = string(k) // m = "x"，类型化字符串常量
const Σ = 1 - 0.7i // 非类型化复数常量
const Δ = Σ + 2.0e-4 // 非类型化复数常量
const φ = iota * 1i - 1 / 1i // 非类型化复数常量
const ic = complex(0, c) // ic = 3.75i，非类型化复数常量
const iθ = complex(0, θ) // iθ = 1i，类型化复数常量
```

在上面这些示例中，阅读时清留意常量的类型化和非类型化的区别。

根据`const`语法定义，在定义常量的时候还可以通过常量列表来进行批量定义，例如：

```go
const a, b, c = 1, 3, "hi" // a = 1，b = 3，c = "hi"，分别是非类型化整型常量和非类型化字符串常量
const u, v float32 = 0, 3 // u = 0.0，v = 3.0，都是类型化单精度浮点常量
```

### iota

在使用带括号的常量声明列表时，用于定义常量的表达式可以从第二个常量（`ConstSpec`）处省略，这样形成的空列表就相当于前面第一个非空表达式列表及其类型。所以省略表达式列表就相当于重复之前的表达式列表。这种用法通常都会结合`iota`这个预定义标识符来使用。

`iota`标识连续的非类型化整数常量，它其实是一个索引，值从`0`开始，可以用来构建一组相关的常量。在每一个常量列表开始的时候，`iota`的索引都会被重置为`0`。

以下是几个利用`iota`定义常量列表的例子。

```go
const (
    a = 1 << iota // a = 1，此时iota = 0
    b  // b = 2，此时iota = 1
    c // c = 4，此时iota = 2
)

const (
    a = 1 << iota // a = 1
    b = 1 << iota // b = 2
    c = 3 // c = 3，此处没有使用iota的值
    d = 1 << iota // d = 8，此时iota = 3
)

const a = iota // a = 0
const b = iota // b = 0，每一个const定义序列开始，iota的索引值就被重置了
```

如果在常量定义列表中，有一些值是不需要的，可以使用`_`来跳过，例如：

```go
const (
    a = iota // a = 0
    b // b = 1
    _ // 空白标识符可以跳过某一个值的获取，此处就跳过了iota = 2
    d // d = 3
)
```

```admonish tip
注意Go中的开放性规则，以上示例中的所有内容都是仅能在当前包中访问的，也就是私有的。如果需要在其他包中可以访问，常量的开头字母必须是大写，例如`const A = 5`。

`const`关键字引领的常量定义是Go语言中的一等公民，可以直接在包中声明。
```

## 变量

变量的主要用途是保存值的存储位置。是的，变量保存的不是值本身，而是值在内存中的位置，这里请不要是指针的概念混淆了。一个变量所能够指代的值是由其类型决定的。

Go 会在编译时对**变量声明**、**函数参数**、**函数返回结果**、**函数声明**、**函数签名**预留命名变量存储空间，而调用内置函数`new`或者获取一个复杂字面量的地址，都会在运行时为变量分配一个存储空间。

变量都是通过`var`关键字声明的，其语法定义如下：

```
VarDecl = "var" ( VarSpec | "(" { VarSpec ";" } ")" ) .
VarSpec = IdentifierList ( Type [ "=" ExpressionList ] | "=" ExpressionList ) .
```

根据`var`关键字的语法定义，以下示例中的变量定义都是合法的。

```go
var i int // 变量i仅声明，自动使用零值初始化
var U, V float64 // 可悲外部访问的变量U、V仅声明，自动使用零值初始化
var k = 0 // 使用赋值表达式初始化变量k，k的类型由提供的值推断
var x, y float64 = -1, -2 // 同时定义双精度浮点类型变量x和y，并对其赋值
var (
    a int
    r, s, t = 2.0, 3, "world" // 利用赋值语句同时定义不同类型的变量
)
```

在上面的示例中，有一些变量在声明的时候没有初始化，此时这个变量是使用其零值进行默认初始化的。所以在声明变量的时候，需要注意每个类型的零值是什么样的。

变量声明中同样也可以使用`_`空白标识符来略过一些赋值过程中不需要的内容。例如在判断一个变量的类型的时：`var _, ok = x.(T)`。

```admonish tip
`var`关键字引领的变量定义是Go语言中的一等公民，可以直接在包中声明。
```

### 简短变量定义

Go 除了可以使用`var`关键字声明和定义变量以外，还提供了一种简短的变量声明方式，这种简短的变量声明方式在 Go 语言代码中被广泛的使用。这种简短的变量声明语法定义如下。

```
ShortVarDecl = IdentifierList ":=" ExpressionList .
```

使用这种简短的变量定义方式，上面的变量定义示例就可以改写成以下样子。

```go
x, y := -1, -2
k := 0
r, s, t := 2.0, 3, "world"
_, ok := x.(T)
```

使用这种简短变量定义时需要注意的是，简短变量定义中不支持书写变量的类型，也就是说变量的类型是完全依靠推断的。而且与使用`var`关键字的变量声明不同，简短变量声明只能够使用在函数内部。

## 作用域

常量和变量的声明都是将一个非空白标识符绑定到常量和变量的过程，实际上这个声明的过程除了可以绑定常量和变量以外，还可以绑定类型、类型参数、函数、标签、包等，这些其他内容的声明会在介绍相应内容的时候进行说明。

```admonish caution
Go中所有声明的限制是同一个非空白标识符在同一个块中不能声明两次，在同一个包的不同文件中也一样。空白标识符因为不会引入绑定，所以它不是一个声明，所以其使用次数不受限制。
```

一个标识符的作用范围是声明它的代码区块范围，Go 语言使用`{}`来定义一个块。那么对于一个标识符来说，其作用域的确定主要依赖于以下几种规则。

1. Go 语言预定义标识符的作用域恒为程序全局。
1. 在函数之外定义的标识符，作用域为包（`package`）。
1. 被导入包的包名作用域是包含导入声明的文件。
1. 方法的接收器、函数参数、函数返回值变量的作用域是函数体。
1. 函数的类型参数的作用域是从方法接收器开始到函数体末尾。
1. 类型参数标识符的作用域是从类型名称开始直到类型定义结束。
1. 函数内部声明的常量和变量的作用域是从其声明结束开始到包含其声明的块的结束。
1. 函数内部声明的类型标识符的作用域是从类型声明开始直到包含其声明的块的结束。

```admonish tip "作用域遮罩"
如果在一个块的内部构建了一个新的块，而且在这个内部块内声明了与外部块中相同的标识符，那么内部块中声明的标识符就将在其作用域范围内屏蔽外部块的同名标识符，这种情况通常被称为作用域遮罩，也是在声明标识符的时候，比较容易被忽略的一个问题。
```
